/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2008 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <config.h>

#include <sys/types.h>	/* time_t */
#include <sys/stat.h>	/* stat() */
#include <ctype.h>		/* isalnum() */
#include <errno.h>		/* errno */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcmp() */
#include <time.h>		/* time() */
#include <unistd.h>		/* stat() */

/* readdir() */
#ifdef HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include "clex.h"
#include "completion.h"

#include "cfg.h"		/* config_num() */
#include "control.h"	/* control_loop() */
#include "edit.h"		/* edit_update() */
#include "history.h"	/* get_hist_entry() */
#include "inout.h"		/* win_waitmsg() */
#include "lex.h"		/* dequote_txt() */
#include "list.h"		/* stat2type() */
#include "sdstring.h"	/* SDSTR() */
#include "sort.h"		/* num_str_compare() */
#include "userdata.h"	/* username_find() */
#include "ustring.h"	/* USTR() */
#include "util.h"		/* emalloc() */

/* internal use only */
#define COMPL_TYPE_PATHCMD	100	/* search $PATH */
#define COMPL_TYPE_USERDIR	101	/* with trailing slash (~name/) */
#define COMPL_TYPE_ENV2		102	/* with trailing curly brace ${env} */

/* environment completion */
extern char **environ;
static SDSTRING *env_var;	/* environment variable names */
static int env_cnt = 0;		/* number of environment variables */

/*
 * commands are stored in several linked lists depending on
 * the first character. It is more effective than sorting.
 */
#define LIST_NR(ch) ((ch) >= 'a' && (ch) <= 'z' ? (ch) - 'a' : 26)
#define LISTS		27
typedef struct cmd {
	SDSTRING cmd;			/* command name */
	struct cmd *next;		/* linked list ptr */
} CMD;
/* PATHDIR holds info about commands in a $PATH directory */
typedef struct {
	const char *dir;		/* PATH directory name */
	time_t timestamp;		/* time of last successfull directory
							   scan, or 0 */
	dev_t device;			/* device/inode from stat() */
	ino_t inode;
	CMD *commands[LISTS];	/* lists of commands in this directory */
} PATHDIR;

static PATHDIR *pd_list;	/* list of PATHDIRs */
static int pd_cnt = 0;		/* number od PATHDIRs in pd_list */

#define QFL_NONE 0
#define QFL_IQ	 1	/* inside quotes '...' or "..." */
#define QFL_MDQ	 2	/* missing closing double quote */
#define QFL_MSQ	 4	/* missing closing single quote */

/* input: completion request data */
static struct {
	CODE type;			/* type of completion - one of COMPL_TYPE_XXX */
	const char *str;	/* string to be completed */
	int strlen;			/* length of 'str' */
	const char *dir;	/* directory to attempt the completion in */
	CODE qflags;		/* quoting issues - one of QFL_XXX */
	int qlevel;			/* quote level for names when inserting (see needs_quote() in edit.c) */
} rq;
/* output: completion results */
static struct {
	FLAG filenames;		/* names are filenames */
	int cnt;			/* number of completion candidates */
	int err;			/* errno value (if cnt is zero) */
	size_t clen;		/* how many characters following 'str'
						   are the same for all candidates */
	char first_char[256];	/* map: first characters following 'str'*/
} compl;
/* output: candidates */
#define CC_LIST	(panel_compl.candidate)
static int cc_alloc = 0;	/* max number of candidates in CC_LIST */
static int cc_max;			/* max number of candidates to present */

static FLAG done;			/* done: completion not possible or
							  full (not partial) name was completed */

extern int errno;

static void
environ_completion_init(void)
{
	int i;
	const char *src;
	size_t len;

	for (env_cnt = 0; environ[env_cnt];)
		env_cnt++;
	if (env_cnt == 0)
		return;	/* no environment ?! */
	env_var = emalloc(env_cnt * sizeof(SDSTRING));

	for (i = 0; i < env_cnt; i++) {
		src = environ[i];
		for (len = 0; src[len] != '=' && src[len] != '\0';)
			len++;
		SD_INIT(env_var[i]);
		sd_copyn(&env_var[i],src,len);
	}
}

static void
command_completion_init(void)
{
	char *path, *p;
	int i, list;

	if ( (path = getenv("PATH")) == 0)
		return;

	/* split PATH to components */
	for (pd_cnt = 1, p = path = estrdup(path); *p; p++)
		if (*p == ':') {
			*p = '\0';
			pd_cnt++;
		}
	pd_list = emalloc(pd_cnt * sizeof(PATHDIR));

	for (i = 0, p = path; i < pd_cnt; i++) {
		pd_list[i].dir = *p ? p : ".";
		pd_list[i].timestamp = 0;
		for (list = 0; list < LISTS; list++)
			pd_list[i].commands[list] = 0;
		while (*p++)
			;
	}
}

void
completion_initialize(void)
{
	command_completion_init();
	environ_completion_init();
	completion_reconfig();
}

void
completion_reconfig(void)
{
	int i;

	if (cc_alloc > 0) {
		for (i = 0; i < cc_alloc; i++)
			sd_reset(&CC_LIST[i].str);
		free(CC_LIST);
	}

	cc_max = cc_alloc = config_num(CFG_C_SIZE);
	if (cc_max == 0)	/* C_PANEL_SIZE = AUTO */
		cc_alloc = 100;	/* enough for every screen */
	CC_LIST = emalloc(cc_alloc * sizeof(COMPL_ENTRY));
	for (i = 0; i < cc_alloc; i++)
		SD_INIT(CC_LIST[i].str);
}

static const char *
code2string(int type, int uppercase)
{
	int i;
	const char *uc;
	static char lc[24];

	switch (type) {
	case COMPL_TYPE_FILE:
		uc = "FILENAME";
		break;
	case COMPL_TYPE_DIR:
		uc = "DIRECTORY NAME";
		break;
	case COMPL_TYPE_PATHCMD:
	case COMPL_TYPE_CMD:
	case COMPL_TYPE_HIST:
		uc = "COMMAND";
		break;
	case COMPL_TYPE_GROUP:
		uc = "GROUP NAME";
		break;
	case COMPL_TYPE_USER:
	case COMPL_TYPE_USERDIR:
		uc = "USER NAME";
		break;
	case COMPL_TYPE_ENV:
	case COMPL_TYPE_ENV2:
		uc = "ENVIRONMENT VARIABLE";
		break;
	default:
		return "";
	}
	if (uppercase)
		return uc;

	for (i = 0; uc[i]; i++)
		lc[i] = tolower((unsigned char)uc[i]);
	lc[i] = '\0';
	return lc;
}

static int
qcmp(const void *e1, const void *e2)
{
	return (config_num(CFG_NUM_SORT) ? num_str_compare : str_compare)(
	  SDSTR(((COMPL_ENTRY *)e1)->str),
	  SDSTR(((COMPL_ENTRY *)e2)->str));
}

/* simplified version of sort_group() found in sort.c */
#define GROUP_DIR       3
#define GROUP_BDEV      4
#define GROUP_CDEV      5
#define GROUP_OTHER     6
#define GROUP_PLAIN     7
static int
sort_group(int type)
{
	if (IS_FT_PLAIN(type))
		return GROUP_PLAIN;
	if (IS_FT_DIR(type))
		return GROUP_DIR;
	if (config_num(CFG_GROUP_FILES) == 2) {
		if (type == FT_DEV_CHAR)
			return GROUP_CDEV;
		if (type == FT_DEV_BLOCK)
			return GROUP_BDEV;
	}
	return GROUP_OTHER;
}

static int
qcmp_group(const void *e1, const void *e2)
{
	int cmp, group1, group2;

	group1 = sort_group(((COMPL_ENTRY *)e1)->file_type);
	group2 = sort_group(((COMPL_ENTRY *)e2)->file_type);
	cmp = group1 - group2;
	if (cmp)
		return cmp;
	return qcmp(e1,e2);
}

void
compl_prepare(void)
{
	panel_compl.pd->cnt = compl.cnt;
	panel_compl.pd->top = panel_compl.pd->min;
	panel_compl.pd->curs = rq.type == COMPL_TYPE_HIST ? 0 : panel_compl.pd->min;
	panel_compl.filenames = compl.filenames;
	panel_compl.description = code2string(rq.type,1);

	if (rq.type != COMPL_TYPE_HIST)
		qsort(CC_LIST,compl.cnt,sizeof(COMPL_ENTRY),
		  (compl.filenames && config_num(CFG_GROUP_FILES)) ? qcmp_group : qcmp);
	else
		win_remark("commands are listed in order of their execution (recent first)");

	panel = panel_compl.pd;
	/* textline inherited from previous mode */
}

static void
register_candidate(const char *cand, int is_link, int file_type,
  const char *aux)
{
	size_t i;
	static const char *cand0;

	if (rq.type == COMPL_TYPE_PATHCMD && IS_FT_EXEC(file_type))
		/* check for duplicates like awk in both /bin and /usr/bin */
		for (i = 0; i < compl.cnt && i < cc_max; i++)
			if (strcmp(SDSTR(CC_LIST[i].str),cand) == 0 &&
			  IS_FT_EXEC(CC_LIST[i].file_type))
				return;

	if (compl.cnt < cc_max) {
		sd_copy(&CC_LIST[compl.cnt].str,cand);
		CC_LIST[compl.cnt].is_link   = is_link;
		CC_LIST[compl.cnt].file_type = file_type;
		CC_LIST[compl.cnt].aux       = aux;
	}

	compl.first_char[cand[rq.strlen] & 0xFF] = 1;

	if (compl.cnt == 0) {
		cand0 = SDSTR(CC_LIST[0].str); /* cand0 = cand; is wrong */
		compl.clen = strlen(cand0) - rq.strlen;
	}
	else
		for (i = 0; i < compl.clen ; i++)
			if (cand[rq.strlen + i] != cand0[rq.strlen + i]) {
				compl.clen = i;
				break;
			}

	compl.cnt++;
}

static void
complete_environ(void)
{
	int i;

	for (i = 0; i < env_cnt; i++)
		if (rq.strlen == 0 || strncmp(SDSTR(env_var[i]),rq.str,rq.strlen) == 0)
			register_candidate(SDSTR(env_var[i]),0,0,environ[i]);
}

static void
complete_history()
{
	int i;
	const HIST_ENTRY *ph;

	for (i = 0; (ph = get_history_entry(i)); i++)
		if (strncmp(USTR(ph->cmd),rq.str,rq.strlen) == 0)
			register_candidate(USTR(ph->cmd),0,0,
			  ph->failed ? "this command failed last time" : 0);
}

static void
complete_username(void)
{
	const char *login, *fullname;

	username_find_init(rq.str,rq.strlen);
	while ( (login = username_find(&fullname)) )
		register_candidate(login,0,0,fullname);
}

static void
complete_groupname(void)
{
	const char *group;

	groupname_find_init(rq.str,rq.strlen);
	while ( (group = groupname_find()) )
		register_candidate(group,0,0,0);
}

static void
complete_file(void)
{
	FLAG is_link;
	CODE type;
	const char *file, *path;
	struct stat st;
	struct dirent *direntry;
	DIR *dd;

	if ( (dd = opendir(rq.dir)) == 0) {
		compl.err = errno;
		return;
	}

	win_waitmsg();
	pathname_set_directory(rq.dir);
	while ( ( direntry = readdir(dd)) ) {
		file = direntry->d_name;
		if (rq.strlen == 0) {
			if (file[0] == '.') {
				if (file[1] == '\0')
					continue;
				if (file[1] == '.' && file[2] == '\0')
					continue;
			}
		}
		else if (strncmp(file,rq.str,rq.strlen))
			continue;

		if (lstat(path = pathname_join(file),&st) < 0)
			continue;		/* file just deleted ? */
		if ( (is_link = S_ISLNK(st.st_mode)) && stat(path,&st) < 0)
			type = FT_NA;
		else
			type = stat2type(st.st_mode,st.st_uid);

		if (rq.type == COMPL_TYPE_DIR && !IS_FT_DIR(type))
			continue;		/* must be a directory */
		if (rq.type == COMPL_TYPE_CMD
		  && !IS_FT_DIR(type) && !IS_FT_EXEC(type))
			continue;		/* must be a directory or executable */
		if (rq.type == COMPL_TYPE_PATHCMD && !IS_FT_EXEC(type))
			continue;		/* must be an executable */

		register_candidate(file,is_link,type,
		  rq.type == COMPL_TYPE_PATHCMD ? rq.dir : 0);
	}
	closedir(dd);
}

static void
pathcmd_refresh(PATHDIR *ppd)
{
	FLAG stat_ok;
	int list;
	const char *file;
	struct dirent *direntry;
	struct stat st;
	DIR *dd;
	CMD *pc;

	/*
	 * fstat(dirfd()) instead of stat() followed by opendir()
	 * would be a cleaner solution, but dirfd() is an extension
	 * not available on some systems
	 */

	stat_ok = stat(ppd->dir,&st) == 0;
	if (stat_ok && st.st_mtime < ppd->timestamp
	  && st.st_dev == ppd->device && st.st_ino == ppd->inode)
		return;

	/* clear all command lists */
	for (list = 0; list < LISTS; list++) {
		while ( (pc = ppd->commands[list]) ) {
			ppd->commands[list] = pc->next;
			sd_reset(&pc->cmd);
			free(pc);
		}
	}

	ppd->timestamp = time(0);
	if (!stat_ok || (dd = opendir(ppd->dir)) == 0) {
		ppd->timestamp = 0;
		return;
	}
	ppd->device = st.st_dev;
	ppd->inode  = st.st_ino;

	win_waitmsg();
	while ( (direntry = readdir(dd)) ) {
		file = direntry->d_name;
		list = LIST_NR(*file);
		pc = emalloc(sizeof(CMD));
		SD_INIT(pc->cmd);
		sd_copy(&pc->cmd,file);
		pc->next = ppd->commands[list];
		ppd->commands[list] = pc;
	}
	closedir(dd);
}

static void
complete_pathcmd(void)
{
	FLAG is_link;
	CODE file_type;
	int i, list;
	const char *file, *path;
	CMD *pc;
	PATHDIR *ppd;
	struct stat st;

	/* include subdirectories of the current directory */
	rq.type = COMPL_TYPE_DIR;
	complete_file();
	rq.type = COMPL_TYPE_PATHCMD;

	list = LIST_NR(*rq.str);
	for (i = 0; i < pd_cnt; i++) {
		ppd = &pd_list[i];
		if (*ppd->dir == '/') {
			/* absolute PATH directories are cached */
			pathcmd_refresh(ppd);
			pathname_set_directory(ppd->dir);
			for (pc = ppd->commands[list]; pc; pc = pc->next) {
				file = SDSTR(pc->cmd);
				if (strncmp(file,rq.str,rq.strlen) != 0)
					continue;
				if (lstat(path = pathname_join(file),&st) < 0)
					continue;
				if ( (is_link = S_ISLNK(st.st_mode))
				  && stat(path,&st) < 0)
					continue;
				file_type = stat2type(st.st_mode,st.st_uid);
				if (!IS_FT_EXEC(file_type))
					continue;
				register_candidate(file,is_link,file_type,ppd->dir);
			}
		}
		else {
			/* relative PATH directories are impossible to cache */
			rq.dir = ppd->dir;
			complete_file();
		}
	}
}

static void
reset_results(void)
{
	int i;

	/* C_PANEL_SIZE = AUTO */
	if (config_num(CFG_C_SIZE) == 0) {
		/*
		 * substract extra lines and leave the bottom
		 * line empty to show there is no need to scroll
		 */
		cc_max = display.panlines + panel_compl.pd->min - 1;
		LIMIT_MAX(cc_max,cc_alloc);
	}

	compl.cnt = 0;
	compl.err = 0;
	compl.filenames = 0;
	for (i = 0; i < 256; i++)
		compl.first_char[i] = 0;
}

static void
complete_it(void)
{
	if (rq.type == COMPL_TYPE_ENV || rq.type == COMPL_TYPE_ENV2)
		complete_environ();
	else if (rq.type == COMPL_TYPE_USER || rq.type == COMPL_TYPE_USERDIR)
		complete_username();
	else if (rq.type == COMPL_TYPE_GROUP)
		complete_groupname();
	else if (rq.type == COMPL_TYPE_HIST)
		complete_history();
	else {
		compl.filenames = 1;
		if (rq.type == COMPL_TYPE_PATHCMD)
			complete_pathcmd();
		else
			/* FILE, DIR, CMD completion */
			complete_file();
	}
}

/* insert char 'ch' if it is not already there */
static void
condinsert(int ch)
{
	if (USTR(textline->line)[textline->curs] == ch)
		textline->curs++;
	else
		edit_nu_insertchar(ch);
}


static void
insert_candidate(int i)
{
	edit_nu_insertstr(SDSTR(CC_LIST[i].str) + rq.strlen,rq.qlevel);

	if ((compl.filenames && IS_FT_DIR(CC_LIST[i].file_type))
	  || rq.type == COMPL_TYPE_USERDIR /* ~user is a directory */ ) {
		done = 0; /* a directory may have subdirectories */
		edit_nu_insertchar('/');
	}
	else {
		done = 1;
		if (rq.type != COMPL_TYPE_HIST) {
			if (rq.type == COMPL_TYPE_ENV2)
				condinsert('}');

			if (rq.qflags == QFL_IQ)
				/* skip over the closing quote */
				textline->curs++;
			else if (rq.qflags == QFL_MSQ)
				edit_nu_insertchar('\'');
			else if (rq.qflags == QFL_MDQ)
				edit_nu_insertchar('\"');

			if (rq.type != COMPL_TYPE_USER)
				condinsert(' ');
		}
	}

	edit_update();
}

static void
show_results(void)
{
	static SDSTRING common = { 0, "" };
	const char *errmsg;

	if (compl.cnt == 0) {
		done = 1;
		switch (compl.err) {
		case 0:
			errmsg = " (no match)";
			break;
		case EACCES:
			errmsg = " (permission denied)";
			break;
		case ENOENT:
			errmsg = " (no such directory)";
			break;
		default:
			errmsg = "";
		}
		win_remark_fmt("cannot complete that %s%s",
		  code2string(rq.type,0),errmsg);
		return;
	}

	if (compl.cnt == 1) {
		insert_candidate(0);
		return;
	}

	if (compl.clen) {
		/* insert the common part of all candidates */
		sd_copyn(&common,SDSTR(CC_LIST[0].str) + rq.strlen,compl.clen);
		edit_insertstr(SDSTR(common),rq.qlevel);
		/*
		 * pretend that the string to be completed already contains
		 * the chars just inserted
		 */
		rq.strlen += compl.clen;
		/* 'compl.first_char' map is no longer valid */
	}

	if (compl.cnt <= cc_max) {
		control_loop(MODE_COMPL);
		return;
	}

	/* show at least the following character */
	win_completion(compl.cnt,compl.clen ? 0 : compl.first_char);
}

static int
is_az09(int ch)
{
	return ch == '_' || (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9');
}

/*
 * compl_file() attempts to complete the partial text in
 * the command line
 *
 * return value:
 *    0  = completion process completed (successfully or not)
 *   -1, -2 = nothing to complete (current word is an empty string):
 *       -1 = first word (usually the command)
 *       -2 = not the first word (usually one of the arguments)
 *   -3 = could complete, but not allowed to (COMPL_TYPE_FILTER)
 */
# define NAMECH(X) ((unsigned char)plinestart[X])

int
compl_file(int type)
{
	static USTRING dequote_str = UNULL, dequote_dir = UNULL;
	const char *lex, *p, *pslash, *pstart, *pend, *plinestart, *plineend;
	int i, start, end, dirlen;
	char ch, chx;
	FLAG first;		/* it is the first word of the command */
	FLAG var;		/* there is $var in the name to be completed */
	FLAG tilde;		/* ther is tilde substitution ~user in the*/
	FLAG type_name;	/* completing a (user|group|environ) name */
					/* assuming character set 0-9A-Za-z_ */

	done = 1;
	if (textline->size == 0)
		return -1;

	/*
	 * plinestart -> start of the input line
	 * plineend   -> trailing null byte of the input line
	 * pstart -> start of the string to be completed
	 * pend   -> position immediately after the last character of the string
	 * pslash -> last slash '/' in the string (if any)
	 */
	plinestart = USTR(textline->line);
	plineend = plinestart + textline->size;

	rq.qflags = QFL_NONE;
	rq.qlevel = 1;
	first = 1;
	var = 0;
	type_name = 0;

	if (type == COMPL_TYPE_DIRPANEL || type == COMPL_TYPE_HIST) {
		/* complete the whole line */
		start = 0;
		end = textline->size;
	}
	else {
		start = end = textline->curs;
		if (type == COMPL_TYPE_USER || type == COMPL_TYPE_GROUP || type == COMPL_TYPE_ENV)
			type_name = 1;
		lex = cmd2lex(plinestart);
		if (IS_LEX_WORD(lex[start]) && (!type_name || is_az09(NAMECH(start)))) {
			/* complete the name at the cursor */
			while (end++, IS_LEX_WORD(lex[end]) && (!type_name || is_az09(NAMECH(end))))
				;
		}
		else if (IS_LEX_WORD(lex[start - 1]) && (!type_name || is_az09(NAMECH(start - 1))))
			/* complete the name immediately before the cursor */ ;
		else if ((IS_LEX_CMDSEP(lex[start - 1]) || IS_LEX_SPACE(lex[start - 1]))
			 && IS_LEX_EMPTY(lex[start])) {
			/* no name to complete */
			while (start--, IS_LEX_SPACE(lex[start]))
				;
			return IS_LEX_CMDSEP(lex[start]) ? -1 : -2;
		} else {
			/* the text at the cursor is not a name */
			win_remark("cannot complete a special symbol");
			return 0;
		}

		if (type == COMPL_TYPE_FILTER)
			return -3;

		while (IS_LEX_WORD(lex[start - 1]) && (!type_name || is_az09(NAMECH(start - 1))))
			start--;

		if (type == COMPL_TYPE_ENV
		  && IS_LEX_WORD(lex[start - 1]) && NAMECH(start - 1) == '{'
		  && IS_LEX_WORD(lex[start - 2]) && NAMECH(start - 2) == '$')
			type = COMPL_TYPE_ENV2;
		else if (type == COMPL_TYPE_AUTO) {
			/* find out what precedes the name to be completed (chx) */
			for (i = start; chx = lex[--i], IS_LEX_SPACE(chx); )
				;
			if (chx == LEX_OTHER) {
				if (i == start - 1) {
					/* there is no space after the special symbol */
					win_remark("cannot complete after a special symbol");
					return 0;
				}
				chx = LEX_SPACE;
				first = 0;
			}
			else
				first = IS_LEX_CMDSEP(chx);

			/* special case - name:value or user@host:/path or name=value */
			if (!first && chx != LEX_IO) {
				for (i = start; i < end; i++) {
					if (!IS_LEX_TEXT(lex[i]))
						break;
					ch = NAMECH(i);
					if (i > start && i < textline->curs && (ch == ':' || ch == '=')) {
						start = i + 1;
						if (start == end)
							return -2;
						break;
					}
					if (ch != '-' && ch != '@' && !is_az09(ch))
						break;
				}
			}
		}

		for (i = start; i < end; i++)
			if (lex[i] == LEX_VAR) {
				var = 1;
				break;
			}

		if (lex[end] == LEX_END_ERR_SQ) {
			rq.qlevel = 0;
			rq.qflags = QFL_MSQ;
		}
		else if (lex[end] == LEX_END_ERR_DQ) {
			rq.qlevel = 2;
			rq.qflags = QFL_MDQ;
		}
		else if (lex[end - 1] == LEX_QMARK) {
			ch = NAMECH(end - 1);
			if (ch == '\'' || ch == '\"')
				rq.qflags = QFL_IQ;
		}
	}

	/* -- at this point we have: 'start' and 'end' --- */

	pstart = plinestart + start;
	pend = plinestart + end;

	pslash = 0;
	if (type != COMPL_TYPE_HIST && !type_name)
		/* separate the name into the directory part and the file part */
		for (p = pend; p > pstart;)
			if (*--p == '/') {
				pslash = p;
				break;
			}

	/* set rq.dir, rq.str */
	if (pslash == 0) {
		rq.str = pstart;
		rq.dir = ".";
	}
	else {
		rq.str = pslash + 1;
		rq.dir = pstart;

		dirlen = pslash - pstart;
		/* dequote 'dir' (if appropriate) + add terminating null byte */
		if (type != COMPL_TYPE_DIRPANEL) {
			tilde = is_dir_tilde(rq.dir);
			dirlen = dequote_txt(&dequote_dir,rq.dir,dirlen);
		}
		else {
			tilde = *rq.dir == '~';
			us_copyn(&dequote_dir,rq.dir,dirlen);
		}
		rq.dir = USTR(dequote_dir);
		if (dirlen == 0)
			/* first slash == last slash */
			rq.dir = "/";
		else if (tilde)
			rq.dir = dir_tilde(rq.dir);
	}

	if (type == COMPL_TYPE_DIRPANEL) {
		if (pslash == 0 && *rq.str == '~') {
			rq.str++;
			rq.type = COMPL_TYPE_USERDIR;
		}
		else
			rq.type = COMPL_TYPE_DIRPANEL;
	}
	else if (type != COMPL_TYPE_AUTO) {
		rq.type = type;
		if (type == COMPL_TYPE_CMD && pslash == 0)
			rq.type = COMPL_TYPE_PATHCMD;
		else if (type == COMPL_TYPE_USER && *rq.str == '~')
			rq.str++;
	}
	/* COMPL_TYPE_AUTO */
	else if (pslash)
		rq.type = first ? COMPL_TYPE_CMD : COMPL_TYPE_FILE;
	else {
		if (*rq.str == '~') {
			rq.str++;
			rq.type = COMPL_TYPE_USERDIR;
		}
		else if (*rq.str == '$' || (*rq.str == '\"' && rq.str[1] == '$' && rq.str++)) {
			if (*++rq.str == '{') {
				rq.str++;
				rq.type = COMPL_TYPE_ENV2;
			}
			else
				rq.type = COMPL_TYPE_ENV;
		}
		else
			rq.type = first ? COMPL_TYPE_PATHCMD : COMPL_TYPE_FILE;
	}

	/* --- after this point there is no TYPE_AUTO --- */

	if (rq.type != COMPL_TYPE_DIR && rq.type != COMPL_TYPE_PATHCMD
	  && rq.type != COMPL_TYPE_CMD && rq.type != COMPL_TYPE_FILE)
		rq.qlevel = 0;
	else if (var) {
		win_remark("cannot complete a name containing a $variable");
		return 0;
	}

	/* dequote 'str' (if appropriate) + add terminating null byte */
	rq.strlen = pend - rq.str;
	if (rq.type != COMPL_TYPE_HIST && rq.type != COMPL_TYPE_DIRPANEL
	  && rq.type != COMPL_TYPE_USERDIR && !type_name) {
		rq.strlen = dequote_txt(&dequote_str,rq.str,rq.strlen);
		rq.str = USTR(dequote_str);
	}
	else if (*pend != '\0') {
		us_copyn(&dequote_str,rq.str,rq.strlen);
		rq.str = USTR(dequote_str);
	}

	if (rq.type == COMPL_TYPE_DIRPANEL)
		rq.type = COMPL_TYPE_DIR;

	/* move cursor to the end of the current word */
	textline->curs = end - (rq.qflags == QFL_IQ /* 1 or 0 */);
	edit_update_cursor();

	done = 0;
	reset_results();
	complete_it();
	show_results();
	return 0;
}

void
cx_compl_complete(void)
{
	insert_candidate(panel_compl.pd->curs);
	next_mode = MODE_SPECIAL_RETURN;
}

static void
complete_type(int type)
{
	if (compl_file(type))
		win_remark("there is nothing to complete");
	if (done && get_current_mode() != MODE_FILE)
		next_mode = MODE_SPECIAL_RETURN;
}

void cx_complete_auto(void)	{ complete_type(COMPL_TYPE_AUTO);	}
void cx_complete_file(void)	{ complete_type(COMPL_TYPE_FILE);	}
void cx_complete_dir(void)	{ complete_type(COMPL_TYPE_DIR);	}
void cx_complete_cmd(void)	{ complete_type(COMPL_TYPE_CMD);	}
void cx_complete_user(void)	{ complete_type(COMPL_TYPE_USER);	}
void cx_complete_group(void){ complete_type(COMPL_TYPE_GROUP);	}
void cx_complete_env(void)	{ complete_type(COMPL_TYPE_ENV);	}
void cx_complete_hist(void)	{ complete_type(COMPL_TYPE_HIST);	}
