/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2008 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <config.h>

#include <sys/types.h>			/* clex.h */
#include <string.h>				/* strlen() */

#include "clex.h"
#include "lex.h"

#include "edit.h"				/* edit_isspecial() */
#include "inout.h"				/* win_warning_fmt() */
#include "ustring.h"			/* us_setize() */

/*
 * analysis of the input line 'cmd'
 * please note that we do not know which shell is going to interpret it
 */
const char *
cmd2lex(const char *cmd)
{
	static USTRING buffer = UNULL;
	char ch, *lex;
	size_t len;
	int i, j, from, to, list;
	FLAG sq, dq, bt;	/* 'single quote', "double quote", `backticks` */
	FLAG inv;

	len = strlen(cmd);
	us_setsize(&buffer,len + 2);
	lex = USTR(buffer) + 1;
	lex[-1] = LEX_BEGIN;
	lex[len] = LEX_END_OK;

	for (i = 0, sq = dq = bt = 0, list = -1; i < len; i++) {
		ch = cmd[i];
		if (sq) {
			if (ch == '\'') {
				lex[i] = LEX_QMARK;
				sq = 0;
			}
			else
				lex[i] = LEX_TEXT;
		}
		else if (dq) {
			if (ch == '\\' && (cmd[i + 1] == '\"' || cmd[i + 1] == '$')) {
				lex[i++] = LEX_QMARK;
				lex[i]   = LEX_TEXT;
			}
			else if (ch == '\"') {
				lex[i] = LEX_QMARK;
				dq = 0;
			}
			else
				lex[i] = ch == '$' ? LEX_VAR : LEX_TEXT;
		}
		else
			switch (ch) {
			case '\\':
				lex[i++] = LEX_QMARK;
				lex[i] = i == len ? LEX_END_ERR_BQ : LEX_TEXT;
				break;
			case '\'':
				lex[i] = LEX_QMARK;
				sq = 1;
				break;
			case '\"':
				lex[i] = LEX_QMARK;
				dq = 1;
				break;
			case '-':
				/* only special inside a list [ ... ] */
				lex[i] = LEX_TEXT2;
				break;
			case '~':
				/* only special in a ~user construct */
				lex[i] = LEX_TEXT2;
				break;
			case '[':
				lex[i] = LEX_TEXT;
				if (list < 0)
					list = i;
				break;
			case ']':
				lex[i] = LEX_TEXT;
				if (list < 0)
					break;	/* no matching [ */
				inv = cmd[list + 1] == '^' || cmd[list + 1] == '!'; /* 0 or 1 */
				if (list + inv == i - 1)
					break;	/* empty [ ] */

				/* valid [ ... ] */
				lex[list++] = LEX_MATCH_BEGIN;
				if (inv)
					lex[list++] = LEX_MATCH_INVERSE;
				for (j = list; j < i; j++)
					if (lex[j] != LEX_QMARK && lex[j] != LEX_TEXT2)
						lex[j] = LEX_MATCH_LIST;
				lex[i] = LEX_MATCH_END;

				/* a-z ranges */
				for (j = list; j < i; j++) {
					if (lex[j] != LEX_TEXT2)
						continue;
					lex[j] = LEX_MATCH_LIST;
					if (cmd[j] != '-')
						continue;

					for (from = j - 1; lex[from] == LEX_QMARK; from--)
						;
					if (lex[from] != LEX_MATCH_LIST)
						continue;
					for (to = j + 1; lex[to] == LEX_QMARK; to++)
						;
					if (lex[to] == LEX_MATCH_END)
						continue;

					lex[from] = LEX_MATCH_RANGE_FROM;
					lex[j]    = LEX_MARK;
					lex[to]   = LEX_MATCH_RANGE_TO;
					j = to;
				}

				list = -1;
				break;
			case '?':
				lex[i] = LEX_MATCH_ONE;
				break;
			case '*':
				lex[i] = LEX_MATCH_ANY;
				break;
			case ' ':
			case '\t':
				lex[i] = LEX_SPACE;
				break;
			case '$':
				lex[i] = LEX_VAR;
				if (cmd[i + 1] == '{')
					lex[++i] = LEX_VAR;
				break;
			case '>':
			case '<':
				lex[i] = LEX_IO;
				break;
			case '&':
			case '|':
			case ';':
			case '(':
				lex[i] = LEX_CMDSEP;
				break;
			case '`':
				lex[i] = TOGGLE(bt) ? LEX_CMDSEP /* opening */
				  : LEX_OTHER /* closing */;
				break;
			default:
				lex[i] = edit_isspecial((unsigned char)ch) ? LEX_OTHER : LEX_TEXT;
		}
	}

	/* open quote is an error */
	if (sq)
		lex[len] = LEX_END_ERR_SQ;
	else if (dq)
		lex[len] = LEX_END_ERR_DQ;

	return lex;
}

/*
 * dequote quoted text: 'src' -> 'dst'
 * return value: the output string length, strlen(dst) <= strlen(src)
 */
int
dequote_txt(USTRING *pustr, const char *src, size_t len)
{
	char ch, *dst;
	size_t i, j;
	FLAG bq, sq, dq;

	us_setsize(pustr,len + 1);
	dst = PUSTR(pustr);

	for (i = j = 0, bq = sq = dq = 0; i < len; i++) {
		ch = src[i];
		if (sq) {
			if (ch == '\'')
				sq = 0;
			else
				dst[j++] = ch;
		}
		else if (dq) {
			if (TCLR(bq)) {
				if (ch != '\"' && ch != '\'' && ch != '$' && ch != '\n')
					dst[j++] = '\\';
				dst[j++] = ch;
			}
			else if (ch == '\\')
				bq = 1;
			else if (ch == '\"')
				dq = 0;
			else
				dst[j++] = ch;
		}
		else if (TCLR(bq))
			dst[j++] = ch;
		else if (ch == '\\')
			bq = 1;
		else if (ch == '\'')
			sq = 1;
		else if (ch == '\"')
			dq = 1;
		else
			dst[j++] = ch;
	}
	dst[j] = '\0';

	return j;
}

void
lex_error(const char *txt, int code)
{
	const char *msg;

	switch (code) {
	case LEX_END_ERR_SQ:
		msg = "missing single quote";
		break;
	case LEX_END_ERR_DQ:
		msg = "missing double quote";
		break;
	case LEX_END_ERR_BQ:
		msg = "trailing backslash";
		break;
	default:
		msg = "error";
	}

	win_remark_fmt("%s: %s",txt,msg);
}
