/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2008 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <config.h>

#include <sys/types.h>	/* pid_t */
#include <errno.h>		/* errno */
#include <fcntl.h>		/* fcntl */
#include <stdio.h>		/* fputs() */
#include <termios.h>	/* struct termios */
#include <unistd.h>		/* STDIN_FILENO */

#include "clex.h"
#include "tty.h"

#include "control.h"	/* err_exit() */

extern int errno;

static struct termios *p_raw, *p_text = 0, *p_save = 0;
#ifdef _POSIX_JOB_CONTROL
static pid_t save_pgid = 0;
#endif

void
tty_initialize(void)
{
	static struct termios text, raw;

	if (tcgetattr(STDIN_FILENO,&text) < 0)
		err_exit("Cannot read the terminal settings");

	raw = text;		/* struct copy */
	raw.c_lflag &= ~(ECHO | ICANON | ISIG | IEXTEN);
	raw.c_cc[VMIN] = 1;
	raw.c_cc[VTIME] = 0;
	p_text = &text;
	p_raw  = &raw;

#ifdef _POSIX_JOB_CONTROL
	/* move CLEX to its own process group */
	save_pgid = tcgetpgrp(STDIN_FILENO);
	setpgid(clex_data.pid,clex_data.pid);
	tcsetpgrp(STDIN_FILENO,clex_data.pid);
#endif
}

void
tty_save()
{
	static struct termios save;

	/* errors are silently ignored */
	p_save = tcgetattr(STDIN_FILENO,&save) == 0 ? &save : 0;
}

void
tty_restore(void)
{
	if (p_save)
		tcsetattr(STDIN_FILENO,TCSAFLUSH,p_save);
}

/* make sure intr = ctrl-C */
/* uses tty_save() */
void
tty_ctrlc()
{
	struct termios ctrlc;

	tty_save();
	if (p_save) {
		ctrlc = *p_save;
		ctrlc.c_cc[VINTR] = CH_CTRL('C');
		tcsetattr(STDIN_FILENO,TCSAFLUSH,&ctrlc);
	}
}

/* noncanonical, no echo */
void
tty_setraw(void)
{
	tcsetattr(STDIN_FILENO,TCSAFLUSH,p_raw);
}

/* note: this is a cleanup function */
void
tty_reset(void)
{
	if (p_text)
		tcsetattr(STDIN_FILENO,TCSAFLUSH,p_text);
}

/* note: this is a cleanup function */
void
tty_pgrp_reset(void)
{
#ifdef _POSIX_JOB_CONTROL
	if (save_pgid)
		tcsetpgrp(STDIN_FILENO,save_pgid);
#endif
}

void
tty_press_enter(void)
{
	int in, flags, loops;

	fputs("Press <enter> to continue. ",stdout);
	fflush(stdout);
	tty_setraw();
	loops = 0;
	while ((in = getchar()) != '\n' && in != '\r') {
		if (in != EOF)
			continue;
		if (errno == EINTR)
			continue;
		if (errno == EAGAIN) {
			flags = fcntl(STDIN_FILENO,F_GETFL);
			if ((flags & O_NONBLOCK) == O_NONBLOCK) {
				/* clear the non-blocking flag */
				fcntl(STDIN_FILENO,F_SETFL,flags & ~O_NONBLOCK);
				continue;
			}
		}
		/* prevent looping in the case of an error */
		if (++loops >= 3)
			err_exit("Cannot read from standard input");
	}
	tty_reset();

	puts("\n----------------------------------------------");
	fflush(stdout);
}
